/*
Bubbles							Copyright 1987 George A. Nelson
								Permission for use granted provided
								this notice remains intact.

By:  George A. Nelson  on	8 April 1987
							16 January 1987 Converted to LightspeedC
							18 Feb 2005 Converted to Codewarrior Pro4
							24 Feb 2005 Converted to Win32 and MSVC6

Description:  Shows pretty bubbles in windows.

*/

#include "..\\Common\\CmnHdr.h"
#include <WindowsX.h>
#include <tchar.h>
#include "main.h"
#include "Lines.h"
#include "Bubbles.h"
#include "..\\Common\\Utility.h"
#include "resource.h"
#include "..\\Common\\HyperLinks.h"


/*	Global Variables:
*/
HINSTANCE hInst;
HWND mainW, clientW;
HMENU gSysMenu, gMenu, gFileMenu, gLinesMenu, gBubblesMenu, gWindowMenu, gHelpMenu;

static const TCHAR *lwcn = _T("LinesWind");
static const TCHAR *bwcn = _T("BubblesWind");

static HMENU gMenuIn = NULL;
static TCHAR gLinesMenuName[32], gBubblesMenuName[32];

#define FIRST_CHILD 100
#define CHILDMENUPROP _T("CHILDMENU")


/*	Message Crackers assume they call something, usually DefWindowProc(), with
	a certain parameter list, but DefFrameProc() has an extra window parameter.
*/
	inline LRESULT CALLBACK
DefFrameProcWrapper( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	return DefFrameProc( hwnd, clientW, uMsg, wParam, lParam );
}


/***************************  About Box  ***************************/

	static LRESULT CALLBACK
About( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam )
{
	switch (message)
	{
	case WM_INITDIALOG:
	{
		CenterWindow( hwnd );

		TCHAR text[4096];
		LoadString( hInst, IDS_ABOUT_TXT, text, chDIMOF(text) );
		HWND txtW = GetDlgItem( hwnd, IDC_TXT );
		SetWindowText( txtW, text );
		
		ConvertStaticToHyperlink( hwnd, IDC_EMAIL );

		/*	Make translucent.  Requires Win2K+.
		*/
		HMODULE user32H = GetModuleHandle( _T("user32.dll") );
		if( user32H ) {
			typedef WINUSERAPI BOOL WINAPI SetLayeredWindowAttributesT(
				HWND hwnd, COLORREF crKey, BYTE bAlpha, DWORD dwFlags);

			SetLayeredWindowAttributesT *SetLayeredWindowAttributesP = (SetLayeredWindowAttributesT *)
				GetProcAddress( user32H, "SetLayeredWindowAttributes" );
			if( SetLayeredWindowAttributesP ) {
				SetWindowLong( hwnd, GWL_EXSTYLE, 
					  GetWindowLong(hwnd, GWL_EXSTYLE) | WS_EX_LAYERED );

				BOOL ok = SetLayeredWindowAttributesP( hwnd, 0, 0xC0, LWA_ALPHA );
				ok = ok;
			}
		}

		return TRUE;
	}

	case WM_COMMAND:
		switch( LOWORD(wParam) )
		{
		case IDOK:
		case IDCANCEL:
			EndDialog(hwnd, LOWORD(wParam));
			return TRUE;

		case IDC_EMAIL:
			if( HIWORD(wParam) == STN_CLICKED ) {
				TCHAR str[256];
				LoadString( hInst, IDS_EMAIL, str, chDIMOF(str) );
				ShellExecute( hwnd, _T("open"), str, _T(""), _T(""), SW_SHOWNORMAL );
			}
			return TRUE;
		}
		break;
	}

    return FALSE;
}


/***************************  Main Wind  ***************************/


	static BOOL
Cls_OnCreate( HWND hwnd, LPCREATESTRUCT lpCreateStruct )
{
	gSysMenu = GetSystemMenu( hwnd, FALSE );
	gMenu = GetMenu( hwnd );
	gFileMenu    = GetSubMenu( gMenu, 0 );
	gLinesMenu   = GetSubMenu( gMenu, 1 );
	gBubblesMenu = GetSubMenu( gMenu, 2 );
	gWindowMenu  = GetSubMenu( gMenu, 3 );
	gHelpMenu    = GetSubMenu( gMenu, 4 );

	GetMenuString( gMenu, 1, gLinesMenuName, chDIMOF(gLinesMenuName), MF_BYPOSITION );
	RemoveMenu( gMenu, 1, MF_BYPOSITION );
	GetMenuString( gMenu, 1, gBubblesMenuName, chDIMOF(gLinesMenuName), MF_BYPOSITION );
	RemoveMenu( gMenu, 1, MF_BYPOSITION );
	DrawMenuBar( hwnd );

	CLIENTCREATESTRUCT cc;
	cc.hWindowMenu = GetSubMenu( gMenu, 1 );
	cc.idFirstChild = FIRST_CHILD;
	clientW = CreateWindow( _T("MDICLIENT"), NULL,
		WS_CHILD | WS_CLIPCHILDREN | WS_VISIBLE | WS_VSCROLL | WS_HSCROLL,
		0, 0, 0, 0, hwnd, (HMENU) 1, hInst, (LPSTR) &cc );
	
	#ifndef _DEBUG
		srand( GetTickCount() );
	#endif

	return true;
}


/*	Make a new MDI Child window.
*/
	static void
MakeWindow( LPCTSTR className, HMENU childMenu )
{
	RECT oR, nR;
	GetClientRect( clientW, &nR );

	HWND cw = (HWND) SendMessage( clientW, WM_MDIGETACTIVE, 0, 0 );
	if( cw ) {
		GetChildRect( cw, clientW, &oR );

		long height = oR.bottom - oR.top;
		long width  = oR.right  - oR.left;
		RECT tR;

		if( height > width ) {
			tR.bottom = height / 2;
			tR.right  = width;
			tR.top    = oR.bottom - tR.bottom;
			tR.left   = oR.left;
			
			oR.bottom = tR.top;
		}
		else {
			tR.bottom = height;
			tR.right  = width / 2;
			tR.top    = oR.top;
			tR.left   = oR.right - tR.right;

			oR.right  = tR.left;
		}

		if( tR.bottom < 40  ||  tR.right < 40 ) {
			cw = NULL;						/* too small to split */
		}
		else {
			MoveWindow( cw, oR.left, oR.top, oR.right-oR.left, oR.bottom-oR.top, TRUE );
			nR = tR;
		}

	}

	MDICREATESTRUCT mcs = { 0 };
	mcs.szClass = className;
	mcs.szTitle = className;
	mcs.hOwner  = hInst;
	mcs.x       = nR.left;
	mcs.y       = nR.top;
	mcs.cx      = nR.right;
	mcs.cy      = nR.bottom;
	mcs.lParam  = (LONG) cw;
	HWND nw = (HWND) SendMessage( clientW, WM_MDICREATE, 0, (LPARAM) &mcs );
	if( nw )
		SetProp( nw, CHILDMENUPROP, childMenu );
}


/*	Keep the proper child's menu in the menu bar.

	-- The way I find the menu position is gross.  The position changes when
	child windows are maximized, anc conceivably at other times.
*/
	static void
CheckMenuBar( HWND hwnd )
{
	HMENU menu = NULL;
	HWND cw = (HWND) SendMessage( clientW, WM_MDIGETACTIVE, 0, 0 );
	if( cw ) {
		menu = (HMENU) GetProp( cw, CHILDMENUPROP );
		if( !IsMenu(menu) )
			menu = NULL;
	}
	if( menu != gMenuIn ) {
		short pos;
		for( pos = 0;  pos < GetMenuItemCount(gMenu);  ++pos ) {
			if( GetSubMenu(gMenu, pos) == gFileMenu )
				break;
		}
		pos++;

		BOOL needDraw = FALSE;
		if( gMenuIn ) {
			RemoveMenu( gMenu, pos, MF_BYPOSITION );
			needDraw = TRUE;
		}
		gMenuIn = menu;
		if( gMenuIn ) {
			TCHAR *name = NULL;
			if( gMenuIn == gLinesMenu )
				name = gLinesMenuName;
			else if( gMenuIn == gBubblesMenu )
				name = gBubblesMenuName;
			InsertMenu( gMenu, pos, MF_BYPOSITION | MF_POPUP, (UINT) gMenuIn, name );
			needDraw = TRUE;
		}
		if( needDraw )
			DrawMenuBar( hwnd );
	}
}


		static BOOL CALLBACK
	sCloseEnumProc( HWND hwnd, LPARAM lParam )
	{
		if( GetWindow(hwnd, GW_OWNER) )
			return 1;
		
		SendMessage( GetParent(hwnd), WM_MDIRESTORE, (WPARAM) hwnd, 0 );
		
		if( !SendMessage(hwnd, WM_QUERYENDSESSION, 0, 0) )
			return 1;
		
		SendMessage( GetParent(hwnd), WM_MDIDESTROY, (WPARAM) hwnd, 0 );
		return 1;
	}

	static void
Cls_OnCommand( HWND hwnd, int id, HWND hwndCtl, UINT codeNotify )
{
	switch( id )
	{
	default:
	{
		HWND cw = (HWND) SendMessage( clientW, WM_MDIGETACTIVE, 0, 0 );
		if( IsWindow(cw) )
			FORWARD_WM_COMMAND( cw, id, hwndCtl, codeNotify, SendMessage );
		FORWARD_WM_COMMAND( hwnd, id, hwndCtl, codeNotify, DefFrameProcWrapper );
		break;
	}

	case ID_FILE_EXIT:
		SendMessage( hwnd, WM_CLOSE, 0, 0 );
		break;

	case ID_HELP_ABOUT:
		DialogBox( hInst, (LPCTSTR) IDD_ABOUTBOX, hwnd, (DLGPROC) About );
		break;

	case ID_FILE_NEWLINES:
		MakeWindow( lwcn, gLinesMenu );
		break;

	case ID_FILE_NEWBUBBLES:
		MakeWindow( bwcn, gBubblesMenu );
		break;

	case ID_FILE_CLOSE:
	{
		HWND cw = (HWND) SendMessage( clientW, WM_MDIGETACTIVE, 0, 0 );
		if( IsWindow(cw)
		 && SendMessage(cw, WM_QUERYENDSESSION, 0, 0) )
		{
			SendMessage( clientW, WM_MDIDESTROY, (WPARAM) cw, 0 );
		}
		break;
	}

	case ID_WINDOW_TILE:
		SendMessage( clientW, WM_MDITILE, 0, 0 );
		break;

	case ID_WINDOW_CASCADE:
		SendMessage( clientW, WM_MDICASCADE, 0, 0 );
		break;

	case ID_WINDOW_ARRANGEICONS:
		SendMessage( clientW, WM_MDIICONARRANGE, 0, 0 );
		break;

	case ID_WINDOW_CLOSEALL:
		EnumChildWindows( clientW, &sCloseEnumProc, 0 );
		break;
	}
}


	static void
Cls_OnInitMenuPopup( HWND hwnd, HMENU hmenu, UINT item, BOOL fSystemMenu )
{
	HWND cw = (HWND) SendMessage( clientW, WM_MDIGETACTIVE, 0, 0 );
	
	if( hmenu == gFileMenu ) {
		DisableMenuItems( hmenu );
		EnableMenuItem( hmenu, ID_FILE_EXIT,       MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_FILE_NEWLINES,   MF_BYCOMMAND | MF_ENABLED );
		EnableMenuItem( hmenu, ID_FILE_NEWBUBBLES, MF_BYCOMMAND | MF_ENABLED );
	}

	if( hmenu == gLinesMenu ||  hmenu == gBubblesMenu )
		DisableMenuItems( hmenu );
	 
	if( hmenu == gHelpMenu ) {
		DisableMenuItems( hmenu );
		EnableMenuItem( hmenu, ID_HELP_ABOUT,      MF_BYCOMMAND | MF_ENABLED );
	}

	if( IsWindow(cw) )
		FORWARD_WM_INITMENUPOPUP( cw, hmenu, item, fSystemMenu, SendMessage );
	FORWARD_WM_INITMENUPOPUP( hwnd, hmenu, item, fSystemMenu, DefFrameProcWrapper );
}


	static void
Cls_OnDestroy( HWND hwnd )
{
	PostQuitMessage( 0 );
}


	static LRESULT CALLBACK
MainWndProc( HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
	HANDLE_MSG( hwnd, WM_CREATE,		Cls_OnCreate	);
	HANDLE_MSG( hwnd, WM_COMMAND,		Cls_OnCommand	);
	HANDLE_MSG( hwnd, WM_INITMENUPOPUP,	Cls_OnInitMenuPopup	);
	HANDLE_MSG( hwnd, WM_DESTROY,		Cls_OnDestroy	);

	case WM_QUERYENDSESSION:
	case WM_CLOSE:
		SendMessage( hwnd, WM_COMMAND, ID_WINDOW_CLOSEALL, 0 );
		if( NULL != GetWindow(clientW, GW_CHILD) )
			return 0;
		break;

	case wmuChildActivate:
		CheckMenuBar( hwnd );
		break;

	case wmuChildClosing:
	{
		HWND cw = (HWND) wParam;
		if( IsWindow(cw) )
			RemoveProp( cw, CHILDMENUPROP );
		CheckMenuBar( hwnd );
	}
	}

	return DefFrameProc( hwnd, clientW, uMsg, wParam, lParam );
}


/*****************************  Main  ******************************/


	int APIENTRY
_tWinMain( HINSTANCE hInstExe, HINSTANCE, LPSTR pszCmdLine, int iCmdShow )	//_txxx from <tchar.h>
{
	hInst = hInstExe;
	
	/*	MainWindProc
	*/
	WNDCLASSEX	mwc = { 0 };
	mwc.cbSize			= sizeof mwc;
	mwc.style			= CS_HREDRAW | CS_VREDRAW;
	mwc.lpfnWndProc		= MainWndProc;
	mwc.hInstance		= hInst;
	mwc.hIcon			= LoadIcon( hInst, MAKEINTRESOURCE(IDI_APPICON) );
	mwc.hCursor			= LoadCursor( NULL, IDC_ARROW );
	mwc.hbrBackground	= (HBRUSH) (COLOR_APPWORKSPACE + 1);
	mwc.lpszMenuName	= (LPCTSTR) MAKEINTRESOURCE( IDR_MENU1 );
	mwc.lpszClassName	= _T("MDISeeThruMainWind");
	mwc.hIconSm			= (HICON) LoadImage( hInst, MAKEINTRESOURCE(IDI_APPICON), IMAGE_ICON,
								 GetSystemMetrics(SM_CXSMICON), GetSystemMetrics(SM_CYSMICON), 0);
	RegisterClassEx( &mwc );

	/*	LinesWindProc
	*/
	WNDCLASSEX	lwc = { 0 };
	lwc.cbSize			= sizeof lwc;
	lwc.style			= CS_HREDRAW | CS_VREDRAW;
	lwc.lpfnWndProc		= LinesWndProc;
	lwc.hInstance		= hInst;
	//lwc.hIcon			= LoadIcon( NULL, IDI_APPLICATION );
	lwc.hCursor			= LoadCursor( NULL, IDC_ARROW );
	lwc.hbrBackground	= GetStockBrush( WHITE_BRUSH );
	lwc.lpszClassName	= lwcn;
	RegisterClassEx( &lwc );

	/*	BubblesWindProc
	*/
	WNDCLASSEX	bwc = { 0 };
	bwc.cbSize			= sizeof bwc;
	bwc.style			= CS_HREDRAW | CS_VREDRAW;
	bwc.lpfnWndProc		= BubblesWndProc;
	//bwc.cbWndExtra		= sizeof( LONG );
	bwc.hInstance		= hInst;
	//bwc.hIcon			= LoadIcon( NULL, IDI_APPLICATION );
	bwc.hCursor			= LoadCursor( NULL, IDC_ARROW );
	bwc.hbrBackground	= GetStockBrush( WHITE_BRUSH );
	bwc.lpszClassName	= bwcn;
	RegisterClassEx( &bwc );

	/*	Load Accelerators
	*/
	HACCEL hAccel = LoadAccelerators( hInst, MAKEINTRESOURCE(IDR_ACCELERATOR1) );

	/*	Create window
	*/
	TCHAR title[64];
	LoadString( hInst, IDS_APP_TITLE, title, chDIMOF(title) );

	mainW = CreateWindow( mwc.lpszClassName, title,
		WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN,
		CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
		NULL, NULL, hInst, NULL );
	ShowWindow( mainW, iCmdShow );

	/*	Parse command line for windows to open.  If arguments, open one Lines and
		one Bubbles window.
	*/
	int argc = __argc;
	LPTSTR *argv = __targv;

	if( argc == 1 ) {
		SendMessage( mainW, WM_COMMAND, ID_FILE_NEWLINES, 0 );
		SendMessage( mainW, WM_COMMAND, ID_FILE_NEWBUBBLES, 0 );

		HWND cw = (HWND) SendMessage( clientW, WM_MDIGETACTIVE, 0, 0 );
		if( cw ) {
			SendMessage( cw, WM_COMMAND, ID_BUBBLES_MAKEBUBBLE, 0 );
			SendMessage( cw, WM_COMMAND, ID_BUBBLES_MAKEBUBBLE, 0 );
			SendMessage( cw, WM_COMMAND, ID_BUBBLES_MAKEBUBBLE, 0 );
			SendMessage( cw, WM_COMMAND, ID_BUBBLES_MAKEBUBBLE, 0 );
			SendMessage( cw, WM_COMMAND, ID_BUBBLES_MAKEBUBBLE, 0 );
		}
	}
	else {
		for( int i = 1;  i < argc;  ++i ) {
			int len = strlen( argv[i] );
			HWND cw = (HWND) SendMessage( clientW, WM_MDIGETACTIVE, 0, 0 );

			if( _tcsncicmp(argv[i], _T("-lines"), len) == 0 )
				SendMessage( mainW, WM_COMMAND, ID_FILE_NEWLINES, 0 );
			else if( _tcsncicmp(argv[i], _T("-bubbles"), len) == 0 )
				SendMessage( mainW, WM_COMMAND, ID_FILE_NEWBUBBLES, 0 );
			else if( _tcsncicmp(argv[i], _T("-next"), len) == 0 ) {
				if( cw )
					SendMessage( clientW, WM_MDINEXT, (WPARAM) cw, 0 );
			}
			else {
				if( cw )
					SendMessage( cw, wmuCmdLineArg, (WPARAM) argv[i], 0 );
			}
		}
	}

	/*	Message loop
	*/
	MSG msg;
	while( GetMessage(&msg, NULL, 0, 0) ) {
		if( !TranslateMDISysAccel(clientW, &msg)
		 && !TranslateAccelerator(mainW, hAccel, &msg) )
		{
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
	}

	return msg.wParam;
}
